/* locolor-demo.c - ANSI Terminal Color Handler */
/* Created: Tue 07 Mar 2023 10:34:32 PM EST malakai */
/* Copyright © 1991-2023 The Last Outpost Project */
/* $Id: locolor-demo.c,v 1.18 2023/05/16 13:42:19 malakai Exp $ */

/* Copyright © 2023 Jeff Jahr <malakai@jeffrika.com>
 *
 * This file is part of Locolor - Last Outpost terminal color library
 *
 * Locolor is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * Locolor is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Locolor.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <math.h>
#include <glib.h>

#include "gvp.h"
#include "locolor.h"
#include "loact.h"

#ifndef ARRAY_SIZE
#define ARRAY_SIZE(array) (sizeof(array) / sizeof(array[0]))
#endif

#ifndef MAX_STRING_LENGTH
#define MAX_STRING_LENGTH 16384
#endif

/* ---- Code Starts Here ---- */

/* NOTE- jump down to demo() for commented example usage of the locolor library. */

/* test code */
void quickcolor(rgba_context *ctx) {

	char buf[MAX_STRING_LENGTH];
	rgba_t color;
	double r,g,b,a=1.0;

	int height = 24;
	int width = 48;

	ctx = resolve_ctx(ctx);

	b=0;
	r=0;
	g=0;
	for(int y=0;y<height;y++) {
		g = (double)y/height;
		for(int x=0;x<width;x++) {
			b = (double)x/width;
			r = (b+g)/2;
			color = rgba_from_double(r,g,b,a);
			// color = rgba_multiply(color,0x33ff66ff);  MAKE IT GREEEEEEEN!
			// sgr_ansi256(buf,sizeof(buf),rgba_greyscale(color,0xFFFFFFFF),ANSI_BG);
			ctx->fgbg_mode = ANSI_BG;
			sgr_ansiColor(buf,sizeof(buf),color,ctx);
			fprintf(stdout,"%s.",buf);
		}
		sgr_ansiNormal(buf,sizeof(buf),ctx);
		fprintf(stdout,"%s|",buf);

		for(int x=0;x<width;x++) {
			b = (double)x/width;
			r = (b+g)/2;
			color = rgba_from_double(r,g,b,a);
			//color = rgba_multiply(color,0x33ff66ff);  MAKE IT GREEEEEEEN!
			//color = rgba_posterize(color,6.0);
			ctx->fgbg_mode = ANSI_BG;
			sgr_ansiColor(buf,sizeof(buf),color,ctx);
			fprintf(stdout,"%s.",buf);
		}

		sgr_ansiNormal(buf,sizeof(buf),ctx);
		fprintf(stdout,"%s\n",buf);
	}
}

/* test code */
int quickpal(char *dst, size_t dlen, rgba_context *ctx) {

	rgba_t color;
	char *d,*eod;

	int idx = 0;
	d=dst;
	eod=dst+dlen;

	ctx = resolve_ctx(ctx);

	d+=sgr_ctxDefaults(d,eod-d,ctx);

	/* The first 16 colors */
	for(int x=0;x<16;x++){
		color = rgba_from_ansi256(idx);
		d+=sgr_ansiColor(d,eod-d,color,ctx);
		d+=g_snprintf(d,eod-d,"%3d",idx);
		idx++;
	}
	d+=sgr_ansiNormal(d,eod-d,ctx);
	d+=g_snprintf(d,eod-d,"\n");

	d+=sgr_ctxDefaults(d,eod-d,ctx);
	/* The color cube */
	for(int y=0;y<6;y++){
		for(int x=0;x<36;x++){
			color = rgba_from_ansi256(idx);
			d+=sgr_ansiColor(d,eod-d,color,ctx);
			d+=g_snprintf(d,eod-d,"%2X",idx);
			idx++;
		}
		d+=sgr_ansiNormal(d,eod-d,ctx);
		d+=g_snprintf(d,eod-d,"\n");
	}

	/* the greyscale */
	d+=sgr_ctxDefaults(d,eod-d,ctx);
	for(int y=0;y<2;y++){
		for(int x=0;x<12;x++){
			color = rgba_from_ansi256(idx);
			d+=sgr_ansiColor(d,eod-d,color,ctx);
			d+=g_snprintf(d,eod-d,"%4d",idx);
			idx++;
		}
		d+=sgr_ansiNormal(d,eod-d,ctx);
		d+=g_snprintf(d,eod-d,"\n");
	}
	return(d-dst);

}

/* test code */
void quickhsv(int width, int height, rgba_context *ctx) {

	rgba_t color;
	hsva_t h;
	char buf[MAX_STRING_LENGTH];
	char fill[6];
	char lorem[]="Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum. ";
	int off = 0;

	h.a = 1.0;
	h.s = 1.0;
	h.v = 1.0;

	ctx = resolve_ctx(ctx);

	strcpy(fill," ");

	for(int y=0;y<height;y++) {
		h.v = unit_clip(2.0*(double)y/height); /* 0-1-1 */
		h.s = unit_clip((1-(2.0*(double)(y-(height/2.0))/height))); /*1-1-0*/
		for(int x=0;x<width;x++) {
			h.h = 360.0 * (double)x/width - 60;
			color = rgba_from_hsva(&h);
			if(ctx->fgbg_mode == ANSI_FG) {
				*fill = *(lorem + ((off++)%(sizeof(lorem)-1)));
			}

			sgr_ansiColor(buf,sizeof(buf),color,ctx);
			fprintf(stdout,"%s%s",buf,fill);
		}
		sgr_ansiNormal(buf,sizeof(buf),ctx);
		fprintf(stdout,"%s\n",buf);
	}

}

/* test code */
void quickblend(int width, int height, rgba_t background, rgba_context *ctx) {

	rgba_t color;
	hsva_t h;
	char buf[MAX_STRING_LENGTH];
	char fill[6];
	char lorem[]="Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum. ";
	int off = 0;

	h.a = 1.0;
	h.s = 1.0;
	h.v = 1.0;
	ctx = resolve_ctx(ctx);

	strcpy(fill," ");

	for(int y=0;y<height;y++) {
		h.a = (double)(height-y)/height;
		/* set the BG at the start of each line. */
		//sgr_ansiColor(buf,sizeof(buf),background,ANSI_BG,cm);
		//fprintf(stdout,"%s",buf);
		for(int x=0;x<width;x++) {
			h.h = 360.0 * (double)x/width - 60;
			*fill = *(lorem + ((off++)%(sizeof(lorem)-1)));
			
			color = rgba_from_hsva(&h);
			color = rgba_gammablend(color,background,ctx->gamma);
			//color = rgba_alphablend(color,background);
			sgr_ansiColor(buf,sizeof(buf),color,ctx);
			fprintf(stdout,"%s%s",buf,fill);
		}
		/* reset the attribs at end of each line. */
		sgr_ansiNormal(buf,sizeof(buf),ctx);
		fprintf(stdout,"%s\n",buf);
	}

}

/* This is the usage demo for locolor.  Read the comments for usage
 * information, run the locolordemo to see the actual output. */
void demo(void) {

	/* --- variable definitions for the demo. ---- */

	/* Most locolor calls require a pointer to a context block.  The context
	 * block holds internal state about the endpoint terminal, default
	 * foreground and background colors, the the ANSI color mode to use with
	 * the terminal, etc.  It's okay pass a NULL pointer as context as the
	 * calls will interpret this as 'use the default context'.  */
	rgba_context *ctx = NULL;

	/* rgba_t is a typedef on a 32bit unsigned integer.  It holds 8 bits of
	 * red, green, blue, and alpha channel information, and is the base storage
	 * type for color data in the locolor module. The fields are RRGGBBAA, so
	 * you can easily write colors as 32bit hex ints in code.*/

	rgba_t color = 0xFFFFFFFF;  /* full rgb white, full alpha opacity. */
	rgba_t red = 0xFF0000FF;    /* This one is red.  */
	rgba_t green = 0x007f00FF;  /* This one is 50% green.  */

	/* Note- ansi terminals can't really do alpha transparency, but keeping the
	 * A channel around makes some color blending operations easier, and would
	 * be compatible with web-based terminals if that support was added in the
	 * future.*/
	/* rgba_t blue = 0x0000FF80;   This one is 50% transparent blue.  */
	rgba_t blue = 0x0000FFFF;   /* Full on Blue.  */
	rgba_t legendary;           /* we'll use this one later. */
	rgba_t maize;               /* we'll use this one later. */

	/* here's a nice fat string buffer for us to play with, along with some
	 * useful pointers. */
	char buf[MAX_STRING_LENGTH];
	 
	 /* a nice generic string pointer... kind of like i for loops. */
	 char *s;	

	 /* shorthand for 'end of s'.  We'll use it later with g_snprintf. */
	 char *eos; 

	 /* some useful default values. */
	 double tempcr;

	/* ---- code starts here! ---- */

	/* the first thing you should do is init the locolor module.  This will
	 * create the default context and default color palette. Should only do
	 * this once.*/
	printf("Init the locolor module.\n");
	locolor_init();

	/* If you've got more than one terminal endpoint to talk to, you can and
	 * should create your own context for each one.  If this was for a MUD type
	 * game, for example, every player's connection should have its very own
	 * rgba_context.*/

	/* if you're writing a standalone app with just one context, you can get
	 * read/write access to the default context like this: */
	/* ctx = resolve_ctx(NULL) */
	/* But for this demo, we'll make our own context. */
	printf("Create a new rgba_context.\n");
	ctx = new_rgba_context();

	/* The new context provides a palette that can contain named color
	 * definitions.  it is initially empty, but we can load new definitions
	 * into it like this: */
	char *userpalette = "locolors.txt";
	printf("Loading user context palette '%s'...",userpalette);
	int entries = load_colorchip_hash(userpalette,ctx);
	printf("found %d entries.\n",entries);

	/* The default context also provides a palette of ANSI colors with standard
	 * names, and the lookup command can fall back and check that palette if
	 * you want it to.  You can control whether or not to fall back on those
	 * colors like this... */
	 
	printf("Checking user palette only...\n");
	ctx->system_palette_fallback = 0;
	if( lookup_color_name("Cyan",ctx,&color) ) {
		/* note that to print an rgba_t color as a web format one, use #%08X. */
		printf("Found that 'Cyan' is #%08X.\n",color);
	} else {
		printf("Couldn't find 'Cyan'.\n");
	}

	printf("Checking with fallback...\n");
	ctx->system_palette_fallback = 1;
	if( lookup_color_name("Cyan",ctx,&color) ) {
		/* note that to print an rgba_t color as a web format one, use #%08X. */
		printf("Found that 'Cyan' is #%08X.\n",color);
	} else {
		printf("Couldn't find 'Cyan'.");
	}

	/* the color definitions file can use aliases for colors too.  In the color
	 * file we're using here, 'Orange' is defined as #FFA500, and 'Legendary'
	 * is defined as 'Oragnge'. */
	printf("Checking aliased color...\n");
	if( lookup_color_name("Legendary",ctx,&color) ) {
		/* note that to print an rgba_t color as a web format one, use #%08X. */
		printf("Found that 'Legendary' is #%08X.\n",color);
		legendary = color;  /* we'll use it later! */
	} else {
		legendary = 0xdeadbeef;
		printf("Couldn't find 'Legendary'.");
	}

	/* the context lets you specifiy the ANSI terminal color level that the
	 * terminal supports.  Figuring what the endpoint terminal actually
	 * supports out is not something that the library does automatically.  If
	 * you were using the library in the context of a MUD game, ANSI_STD is
	 * supported by almost every terminal.  If you detect that that the
	 * terminal has 256 color support, like via a TERMTYPE, MTTS, or MNES
	 * negotiation, the ANSI_240 mode is a good option.  It avoids using the
	 * first 16 colors, which many terminals remap to custom colors.  If the
	 * enpoint supports 24 bit color, you can use ANSI_TRUECOLOR and get very
	 * fine grained control, or if you don't want any color at all, use
	 * ANSI_MONO. See ansi_color_mode in locolor.h for more options.*/

	/* color_mode_from_env() will return ANSI_240, or ANSI_TRUECOLOR if the
	 * COLORTERM env varibale is set to 'truecolor'. */
	ctx->color_mode = color_mode_from_env();

	/* After setting the color_model value, you can always talk about color in
	 * terms of rgba_t, and let the library worry about translating the 32 bit
	 * value into the closest color that the terminal can display.  This is
	 * accomplished by sending the correct "Set Graphic Rendition" codes down
	 * to the terminal.  The library provides a generic function for adding the
	 * right sgr terminal sequence into a buffer using the sgr_ansiColor()
	 * function.  sgr_ansiColor() takes care of respecting the color mode,
	 * gamma blending, contrast ratio, and supressing redundant color commands.
	 * The other sgr_*() functions do not. */
	
	/* NOTE- I like using scnprintf() from the Linux Kernel or g_snprintf()
	 * from GLib for printing strings into buffers.  They have a pointer and a
	 * length, and most importantly return the number of char's that were
	 * actually printed into the buffer.  If you try to print past the end of
	 * the buffer, it won't do it, and it won't lie about how many characters
	 * 'would have been printed' like plain old snprintf() does.  Watch how s
	 * and eos works, this demo uses it a lot.*/

	 s = buf; /* point s to the start of the buffer */
	 eos = buf + sizeof(buf);  /* point eos to the end of the buffer. */

	 /* s tells g_snprintf where to start printing, eos-s is the maximum length
	  * it can print into.  on return, increment s by the number of bytes
	  * written.  A subsequent call will then automatically append into buf
	  * starting at the new s. */
	 s += g_snprintf(s,eos-s,"Testing out sgr_ansiColor() ");
	 s += g_snprintf(s,eos-s,"using g_snprintf()...");
	 /* now you can puts() the buf, or send_to_char() it, or whatever. */
	 puts(buf); 

	/* sgr_* functions work like g_snprintf() does.  You can use them to append
	 * terminal sequences in the same way.  There were some static colors
	 * defined up above, lets try them out by sending some color change
	 * sequences. */

	printf("Building up a test string.\n");

	/* reset s to the start of the buf. (eos is still pointing to the end, no
	 * need to change it.) */
	s = buf;  

	/* This sgr_ctxDefaults() call will reset the graphic rendition state, then
	 * set the default forground and background color. */
	s+=sgr_ctxDefaults(s,eos-s,ctx);
	s+=g_snprintf(s,eos-s,"Some default text. ");
	/* look up above... we defined red up there. */
	s+=sgr_ansiColor(s,eos-s,red,ctx);
	s+=g_snprintf(s,eos-s,"This is some red text. ");
	s+=sgr_ansiColor(s,eos-s,legendary,ctx);
	s+=g_snprintf(s,eos-s,"This is some legendary text.");
	s+=sgr_ansiNormal(s,eos-s,ctx);  /* lets put it back to normal. */
	s+=g_snprintf(s,eos-s,"Some normal text. ");
	puts(buf);

	/* The loact.h library calls add locolor_act() and locolor_printf()
	 * routines that make it a lot easier to write strings with color.  You
	 * have to init the module with TRUE if you want to use the builtin parser
	 * commands, or FALSE if you want to leave them unregistered. */
	printf("Init the locolor_act module.\n");
	locolor_act_init(TRUE);

	/* The library provides an example of a Set Graphic Rendition substitution
	 * parser, sgr_subtitution().  Like the other sgr_*() calls, it works like
	 * g_snprintf(). This should print out the same thing as above.*/
	s=buf;
	s+=locolor_act(
		"${reset}Some normal text. ${fg(#FF0000)}This is some red text. ${fg(Legendary)}This is some legendary text.$reset",
		s,eos-s,ctx
	);
	puts(buf);

	/* It would be nice to not have to deal with all of those s and buf
	 * pointers all the time though.  To keep the rest of the demo and test
	 * code clean, the library provides a wrapper around printf and
	 * locolor_act, so that we can use SGR commands directly in a
	 * printf-like thing.  That thing is called locolor_printf(ctx,"",...) and
	 * we'll use it for a lot of the tests in this demo. */

	locolor_printf(ctx,
		"locolor_printf() test: %d: $bg(Blue)printf()$bg gets called first.  %d: $bg(%s)then${reset} locolor_act.\n",
		1,2,"#FF0000"
	);

	locolor_printf(ctx,
		"locolor_act fgbg: Both fg and bg can change.  $fgbg(Maize:Blue)Hail to the victors!${reset}\n"
	);

	/* Note that empty args to $fg $bg or $fgbg will default to the context
	 * defaults. */
	locolor_printf(ctx,
		"$fgbg()Empty args test: $fgbg(Maize:Blue)Hail to the victors!$fgbg() Hooray!\n"
	);

	/* change the defaults, but remember them so we can put them back. */
	rgba_t fgtmp = ctx->default_style.fg;
	rgba_t bgtmp = ctx->default_style.bg;
	ctx->default_style.fg = 0xFFFFFFFF; /* bright white */
	ctx->default_style.bg = 0x18453bff; /* MSUGreen */

	locolor_printf(ctx,
		"Empty args test: setting default_fg=#%08X default_bg=#%08X.\n",
		ctx->default_style.fg,
		ctx->default_style.bg
	);

	locolor_printf(ctx,
		"Empty args test: $fgbg()Smash right through that line of blue! $fgbg(Maize:Blue)Hail to the victors!$fgbg() Hooray! ${reset} Rah Rah!\n"
	);
	locolor_printf(ctx,
		"Empty args test: Sending a default${default} resets the terminal, and reasserts fg and bg colors.${reset}\n"
	);

	/* done with that test, put 'em back. */
	ctx->default_style.fg = fgtmp;
	ctx->default_style.bg = bgtmp;
	locolor_printf(ctx,
		"Empty args test: setting default_fg=#%08X default_bg=#%08X.\n",
		ctx->default_style.fg,
		ctx->default_style.bg
	);
	locolor_printf(ctx,
		"Empty args test: $fg(Yellow)Sending a reset${reset} puts the terminal back to its local defaults.\n"
	);
	
	/* the contrast ratio can be set in the context to control the legibility
	 * of text. w3c recommends a minimum contrast ratio of 4.5.*/
	puts("Contrast Test!");
	double tmpcr = ctx->contrast_ratio;
	for(double cr=1.0;cr<=4.5;cr+=0.5) {
		ctx->contrast_ratio = cr;
		locolor_printf(ctx,
			"Contrast Ratio = %2.2f: "
			"$bg(Purple)Can this $fg(Green)green text on a $fg(Purple)Purple$fg(Green) background can be read?${reset}\n",
			ctx->contrast_ratio
		);
	}
	ctx->contrast_ratio = tmpcr;

	/* the A channel in a fg change blends against the current bg color.*/
	puts("Opacity, Gamma, Contrast Ratio Test!");
	ctx->fgbg_mode = ANSI_FG;
	for(int i=10;i>0;i--) {
		color = rgba_alpha(0x00FF00FF,i*0xFF/10);
		ctx->contrast_ratio = 1.0;
		ctx->gamma = 1.0;
		locolor_printf(ctx,"#%08X: $fg(#%08X)Gamma=%3.2f cr=%3.2f, ",
			color,color,ctx->gamma,ctx->contrast_ratio
		);
		ctx->gamma = 2.2;
		locolor_printf(ctx,"$fg(#%08X)Gamma=%3.2f cr=%3.2f, ",
			color,ctx->gamma,ctx->contrast_ratio
		);
		ctx->contrast_ratio = 2.5;
		locolor_printf(ctx,"$fg(#%08X)Gamma=%3.2f cr=%3.2f, ${reset}#%08X\n",
			color,ctx->gamma,ctx->contrast_ratio,
			color
		);
	}
	ctx->contrast_ratio = 1.0;

	/* Now lets look at some different output modes. quickpal() prints out an
	 * ansi256 color style palette using the sgr_ansiColor() function, and it
	 * respects the context settings. Lets try some options.*/
	ansi_color_mode cmdefault = ctx->color_mode;
	ansi_color_mode showcm[] = { ANSI_MONO, ANSI_STD, ANSI_240, ANSI_TRUECOLOR };

	for(int i=0;i<ARRAY_SIZE(showcm)&&(cmdefault>=showcm[i]);i++) {
		s=buf;
		ctx->color_mode = showcm[i];
		s+=g_snprintf(s,eos-s,"Color mode %d:\n",ctx->color_mode);
		s+=quickpal(s,eos-s,ctx);
		s+=g_snprintf(s,eos-s,"(That took %zd bytes)\n",(s-buf));
		puts(buf);
	}
	ctx->color_mode = cmdefault;

	/* The library has a text function for doing smooth color blends across
	 * text.  It is another function that works like g_snprintf. It starts the
	 * blend with the current fg and bg colors that are on the terminal as
	 * indictated by the ctx, and modififes them smoothly with each character
	 * printed to the new fg and bg supplied to the call. */
	
	/* Maize is a distinct shade of Yellow, btw.   Blue is... well.. blue.*/
	lookup_color_name("Maize",ctx,&maize); s=buf;
	lookup_color_name("Blue",ctx,&blue); s=buf;
	
	/* set s to the start of the buffer. */
	s = buf; 
	s += g_snprintf(s,eos-s,"Blend FG Test: ");
	/* set the foreground to Maize. */
	s += sgr_ansiColor(s,eos-s,maize,ctx);
	/*put the output back to normal. */
	s+=sgr_ansiNormal(s,eos-s,ctx);
	/* and print it all out. */
	puts(buf);
	
	/* now lets do it the easy way. */
	locolor_printf(ctx,
		"Blend FG test: $fg(Maize)$blend('Hail to the victors!',Blue).${reset}\n"
	);

	/* $blend() takes three arguments: the text, the final forground color, and
	 * the final background color.  If you leave off one or more of the colors,
	 * the default colors are used. */
	locolor_printf(ctx,
		"Blend BG test: $bg(Maize)$blend('Hail to the victors!',,Blue).${reset}\n"
	);

	/* $blend() lets you change foreground and background simultaneously. */
	locolor_printf(ctx,
		"Blend FGBG test: $fgbg(Maize,Blue)$blend('Hail to the victors!',Blue,Maize).${reset}\n"
	);

/* Colors can be specified in different formats, like this. */
	locolor_printf(ctx,
		"Formats: $fg(Legendary)Legendary $fg(Blue)|"
		"$fg(Orange)Orange $fg(Blue)|"
		"$fg(#FFA500FF)#FFA500FF $fg(Blue)|"
		"$fg(#ffa500ff)#ffa500ff $fg(Blue)|"
		"$fg(#FFA500)#FFA500 $fg(Blue)|"
		"$fg(#FA0)#FA0 $fg(Blue)|"
		"$fg(#FA0F)#FA0F"
		"${reset}\n"
	);

	
	/* some minimum contrast ratio examples and tests.  A good minimum contrast
	 * ratio can blends easier to read. */
	char *cr_test_strings[] = { 
		"$fg(Red)$bg(Red)Can $bg(Green)it be $bg(Blue)read?",
		"$fg(Red)Red on default. $bg(Red)Red on red.$bg(Black)  Red on Black.",
		"$fgbg(Blue,Black)$blend('This is some hard to read text that gets even tougher to read.',Blue,Blue)",
		"$fgbg(#FFF,Black)$blend('this is also some hard to read text on a grey background.',#FFF,#FFF)",
		"$fgbg(Maize,Blue)$blend('Hail to the victors!',Blue,Maize)",
		"$fgbg(MSUGreen,BrightWhite)$blend('Smash right through that line of blue',BrightWhite,MSUGreen)",
		NULL
	};
	double contrast_ratios[]= { 1.0, 1.5, 4.3, -1.0 };

	tempcr = ctx->contrast_ratio;
	for(int t=0;cr_test_strings[t];t++) {
		for(int cr=0;contrast_ratios[cr]>=1.0;cr++) {
			ctx->contrast_ratio = contrast_ratios[cr];
			locolor_printf(ctx,"CR Transition test: cr=%3.2f ",ctx->contrast_ratio);
			locolor_printf(ctx,cr_test_strings[t]);
			locolor_printf(ctx,"${reset}\n");
		}
	}
	ctx->contrast_ratio = tempcr;

	/* You can simulate a hue blend with multiple blends in a row. */
	/* (also note the escaped single quote marks in M&M's and doesn't ) */
	locolor_printf(ctx,
		"Fake Rainbow: $fg(Red)Skittles $blend('beat M&M\\'s ',Orange)$blend('hands down! ',Yellow)$blend('Light doesn\\'t melt ',Green)$blend('in your hand ',Blue)$blend('either. ',Indigo)$blend('Taste the ',Violet)rainbow$blend('!!!!!!!!',DefaultFG)${reset}\n"
	);

	/* or use the rainbow blender.  args are
	 * (text,finalfg,finalbg,directionmode).  It works by 'spinning' the hue
	 * clockwise, or counterclockwise depending on the last argument.*/
	
	/* also note the use of ` as a quote character, so that the '  in M&M's
	 * doesn't need escaping. */
	locolor_printf(ctx,
		"Rainbow text red: "
		"$rainbow(`Skittles beat M&M's hands down! Light doesn't melt in your hand either. Taste the rainbow!!!!!!!!`,'Red','Red')"
		"${reset}\n"
	);
	locolor_printf(ctx,
		"Rainbow text  /blue: "
		"$rainbow(`Skittles beat M&M's hands down! Light doesn't melt in your hand either. Taste the rainbow!!!!!!!!`,'/Blue','/Blue')"
		"${reset}\n"
	);
	locolor_printf(ctx,
		"Rainbow text red/blue: "
		"$fg(Red)$rainbow(`Skittles beat M&M's hands down! Light doesn't melt in your hand either. Taste the rainbow!!!!!!!!`,'Red/Blue','Red/Blue')"
		"${reset}\n"
	);
	tmpcr = ctx->contrast_ratio;
	ctx->contrast_ratio = 4.3;
	locolor_printf(ctx,
		"Rainbow text red/blue high contrast: "
		"$fg(Red)$rainbow(`Skittles beat M&M's hands down! Light doesn't melt in your hand either. Taste the rainbow!!!!!!!!`,'Red/Blue','Red/Blue')"
		"${reset}\n"
	);
	locolor_printf(ctx,
		"Rainbow text Green/Green high contrast: "
		"$fg(Red)$rainbow(`Skittles beat M&M's hands down! Light doesn't melt in your hand either. Taste the rainbow!!!!!!!!`,'Green/Green','Green/Green')"
		"${reset}\n"
	);
	ctx->contrast_ratio = 1.0;

	/* Here is the orange text that started this whole adventure! */
	locolor_printf(ctx,
		"Orange Test: $blend(\"I can't believe I'm seeing\",Orange) orange $blend('text on my terminal!',DefaultFG)${reset}\n"
	);

	/* playing with default colors and the special DefaultFG/DefaultBG color names. */
	/* Also... here's an easy way to make a temporary holder for a ctx.  */
	rgba_context tmpctx;
	tmpctx = *ctx;

	ctx->default_style.fg = red;
	ctx->default_style.bg = green;
	locolor_printf(ctx,
		"DefaultFG/BG: $fgbg(DefaultFG,DefaultBG)fg=#%08X bg=#%08X",
		ctx->default_style.fg,ctx->default_style.bg
	);
	/* restore the context to what it was orginally. */
	*ctx = tmpctx;
	locolor_printf(ctx,"${reset}\n");

	locolor_printf(ctx,
		"DefaultFG/BG: ${reset}fg=#%08X bg=#%08X\n",
		ctx->default_style.fg,ctx->default_style.bg
	);
	

	/* check that the blend function works correctly with UTF-8 sequences. */
	locolor_printf(ctx,
		"UTF-8 Blend: $blend('Ï çåῃ ēąԷ ǧĺǟṡš, ἱʈ ďὃęș ᾐȭե ĥǔɼẗ ḿѐ.',Indigo)"
		"$blend('ᛁᚳ᛫ᛗᚨᚷ᛫ᚷᛚᚨᛋ᛫ᛖᚩᛏᚪᚾ᛫ᚩᚾᛞ᛫ᚻᛁᛏ᛫ᚾᛖ᛫ᚻᛖᚪᚱᛗᛁᚪᚧ᛫ᛗᛖ᛬',Orange)"
		"${reset}\n"
	);

	/* check that the blend function works correctly with UTF-8 sequences. */
	locolor_printf(ctx,
		"Some animals: $rainbow(`Noah's ark 🐀 🐁 🐂 🐃 🐄. What a 🌈🐀🕳️ this turned into!`,'BrightRed/BrightBlue','Red/Blue')"
		"${reset}\n"
	);

	/* Thanks for the nyan cat idea, crb. */
	locolor_printf(ctx,"Nyan cat: ");
	locolor_printf(ctx,"$nyan('A test of the builtin var parser. Nyan Cat.')${reset}\n");

	/* The char attributes can be added like this. */
	locolor_printf(ctx,"Italics test: ${italic}'Lorem ipsum dolor sit amet'${italic} is latin for $italic()'BLAH BLAH BLAH'$italic(), see?\n");
	locolor_printf(ctx,"Italics test: ${italic}'Lorem ipsum dolor sit amet'${reset} is latin for ${italic}'BLAH BLAH BLAH'${reset}, see?\n");
	locolor_printf(ctx,"Italics test: $italic(on)'Lorem ipsum dolor sit amet'$italic(off) is latin for $italic(1)'BLAH BLAH BLAH'${italic}, see?\n");
	locolor_printf(ctx,"Bold test: Now that's a ${bold}bold move${bold} right there!\n");
	locolor_printf(ctx,"Faint test: $bold()Its vapors!$bold I'm feeling faint! $faint()Someone catch me!$faint()\n");
	locolor_printf(ctx,"Underline test: This is ${underline}NOT${underline} a hyperlink.\n");
	locolor_printf(ctx,"DoubleUL test: This is $doubleul()NOT a hyperlink$doubleul() either!\n");
	// locolor_printf(ctx,"Blink test: $blink('Blink test was dropped') to defend against weeping angels.\n");
	locolor_printf(ctx,"Reverse test: Put the car into R and $reverse()back up slowly$reverse please.\n");
	locolor_printf(ctx,"Crossedout test: This is $crossedout()NOT a broken hyperlink$crossedout() either.\n");
	locolor_printf(ctx,"Combo underline test: This $underline()Under$crossedout()Cross$doubleul()Double $underline()Under$crossedout()Cross$doubleul()Double.${reset}\n");

	/* these tests show some special handling of the parser. */
	locolor_printf(ctx,"Parser cancel test: $fg(#f00)Red${reset},$- $fg(#fff)White, and $fg(#00F)Blue.${reset}\n");
	locolor_printf(ctx,"Parser double dollar: Single dollar sign $$. $#@#@@ $(What? $CompletelyUnhandled(variable name!) ${italic}$fg(Orange)Cool.${reset}\n");
	locolor_printf(ctx,"Parser unhandled: $#@#@@ $(What? $CompletelyUnhandled(variable name!) ${italic}$fg(Orange)Cool.${reset}\n");

	/* some other act tests */
	locolor_printf(ctx,"rgba_act: Orange is $rgba(Orange), #f0f is $rgba(#f0f), unknown is $rgba(unknown).${reset}\n");
	locolor_printf(ctx,"hsva_act: hsva(128.7,0.35,0.78,1) is $hsva(128.7,0.35,0.78,1) ${reset}\n");

	/* locolor_act can work recursively on arguments.*/
	locolor_printf(ctx,"act recursion: hsva(%s) is $hsva(%s) $fg($hsva(%s))this lovely shade of chartruse.${reset}\n",
		"90.146341,1.0,0.803921,1",
		"90.146341,1.0,0.803921,1",
		"90.146341,1.0,0.803921,1"
	);
	/* args can be quoted srings, or other variables. */
	locolor_printf(ctx,"act recursion: $bold()This is a $reverse()test$reverse() of $italic()embeded $underline()attributes$underline()$italic()$bold().${reset}\n");


	/* remember that you're responsible for throwing out anything you've
	 * allocated! */
	printf("\nRemember to clean up your mess!\n");
	free_rgba_context(ctx);
	locolor_act_quit();
	locolor_quit();
	printf("(Yes, mom.)\n");
}

void quickblends(void) {

	/* --- variable definitions for the demo. ---- */

	rgba_context *ctx = NULL;
	/* ---- code starts here! ---- */

	/* the first thing you should do is init the locolor module.  This will
	 * create the default context and default color palette. Should only do
	 * this once.*/
	printf("Init the locolor module.\n");
	locolor_init();
	locolor_act_init(TRUE); /* bring in the default commands. */

	/* if you're writing a standalone app with just one context, you can get
	 * read/write access to the default context like this: */
	ctx = resolve_ctx(NULL);
	ctx->color_mode = color_mode_from_env();

	/* These are some quick test demos that I cooked up while writing this mess. */
	
	/* quickhsv was nice for testing that different terminal color models were
	 * working.. */

	locolor_printf(ctx,"$fg(Yellow)Quick HSV FG Test.${reset}\n");
	ctx->fgbg_mode = ANSI_FG;
	quickhsv(40,20,ctx);
	locolor_printf(ctx,"$fg(Yellow)Quick HSV BG Test.${reset}\n");
	ctx->fgbg_mode = ANSI_BG;
	quickhsv(40,20,ctx);


	/* This just looks cool. */

	locolor_printf(ctx,"$fg(Yellow)Quick Blend BG Test.${reset}\n");
	ctx->fgbg_mode = ANSI_BG;
	quickblend(40,20,0x00000000,ctx);

	/* remember that you're responsible for throwing out anything you've
	 * allocated! */
	locolor_act_quit();
	locolor_quit();
}

void styledemo(void) {

	/* --- variable definitions for the demo. ---- */

	rgba_context *ctx = NULL;

	char buf[MAX_STRING_LENGTH];
	char *s=buf;	
	char *eos=buf+sizeof(buf); 

	char *userpalette = "locolors.txt";
	char *userstyles = "lostyles.txt";
	int entries;
	style_t *style;

	/* ---- code starts here! ---- */

	locolor_init();
	locolor_act_init(TRUE);
	ctx = new_rgba_context();
	ctx->color_mode = color_mode_from_env();
	//ctx->debug = 1;

	/* The new context provides a palette that can contain named color
	 * definitions.  it is initially empty, but we can load new definitions
	 * into it like this: */
	printf("Loading user context palette '%s'...",userpalette);
	entries = load_colorchip_hash(userpalette,ctx);
	printf("found %d entries.\n",entries);

	printf("Loading user context styles '%s'...",userstyles);
	entries = load_style_hash(userstyles,ctx);
	printf("found %d entries.\n",entries);

	/* set s to the start of the buffer. */
	s = buf; 
	s += g_snprintf(s,eos-s,"Style Test: ");
	/* fetch the style */	
	if(!(style=lookup_style_name("UofM",ctx))) {
		s += g_snprintf(s,eos-s,"Style not found!");
	} else {
		s+=sgr_ansiStyle(s,eos-s,style,ctx);
		s += g_snprintf(s,eos-s,"Hail to the victors!");
		/*put the output back to normal. */
		s+=sgr_ansiNormal(s,eos-s,ctx);
	}
	/* and print it all out. */
	puts(buf);

	/* Now the easy way. */
	locolor_printf(ctx,"Style act test: $style(UofM)Hail to the victors!$style() and such.${style}\n");

	locolor_printf(ctx,"Style act test: $style(Maize/Blue/u)Hail to the victors!$style() and such${style}.\n");

	locolor_printf(ctx,"Style act test: $style(MSU)Victory for MSU!$reset $style(Maize/Blue/u)Hail to the victors!$reset and such.\n");

	locolor_printf(ctx,"Style act test: $style(tableheader)Current Players$reset\n");

	locolor_printf(ctx,"Style act test: Try typing '${style(command)}what${reset}' for info.\n");

	locolor_printf(ctx,"Style act test: $style(Amaizeing)This text should look blue and yellow.$style\n");

	locolor_printf(ctx,"Style act test: $style(Amazing)This text should look sickly?$style\n");

	locolor_printf(ctx,"Style defined test: $style(White/Blue/i)White/Blue/i to $style(Green)Green $style\n");
	locolor_printf(ctx,"Style defined test: $style(White/Blue/i)White/Blue/i to $style(/Green)/Green$style\n");
	locolor_printf(ctx,"Style defined test: $style(White/Blue/i)White/Blue/i to $style(//u)//u$style\n");
	locolor_printf(ctx,"Style defined test: $style(White/Blue/i)White/Blue/i to $style(/)/$style\n");
	locolor_printf(ctx,"Style defined test: $style(White/Blue/i)White/Blue/i to $style(//)//$style\n");
	locolor_printf(ctx,"Style defined test: $style(White/Blue/i)White/Blue/i to $style(Default)//$style\n");
	locolor_printf(ctx,"Style defined test: $style(White/Blue/i)White/Blue/i to $style(Maize)Maize $style\n");


	/* remember that you're responsible for throwing out anything you've
	 * allocated! */
	free_rgba_context(ctx);
	locolor_act_quit();
	locolor_quit();
}

void attrdemo(void) {

	/* --- variable definitions for the demo. ---- */

	rgba_context *ctx = NULL;

	char buf[MAX_STRING_LENGTH];
	char *s=buf;	
	char *eos=buf+sizeof(buf); 

	char *userpalette = "locolors.txt";
	char *userstyles = "lostyles.txt";
	int entries;
	style_t style;

	/* ---- code starts here! ---- */

	locolor_init();
	locolor_act_init(TRUE);
	ctx = new_rgba_context();
	ctx->color_mode = color_mode_from_env();

	for(int attr=0;attr<ANSI_CA_MAX;attr++) {

		/* set s to the start of the buffer. */
		s = buf; 
		s += g_snprintf(s,eos-s,"Attribute Test: %15s ",ansi_attributes[attr].name);
		s += sgr_ansiAttr(s,eos-s,attr,TRUE,ctx);
		s += g_snprintf(s,eos-s,"Lorem ipsum dolor");
		/*put the output back to normal. */
		s += sgr_ansiAttr(s,eos-s,attr,FALSE,ctx);
		s += g_snprintf(s,eos-s," sit amet");
		/* and print it all out. */
		puts(buf);
	}

	/* remember that you're responsible for throwing out anything you've
	 * allocated! */
	free_rgba_context(ctx);
	locolor_act_quit();
	locolor_quit();
}

void blendstyledemo(void) {

	/* --- variable definitions for the demo. ---- */

	rgba_context *ctx = NULL;

	char buf[MAX_STRING_LENGTH];
	char *s=buf;	
	char *eos=buf+sizeof(buf); 

	char *userpalette = "locolors.txt";
	char *userstyles = "lostyles.txt";
	int entries;
	style_t *style;
	GList *sl;
	style_t *st;

	/* ---- code starts here! ---- */

	locolor_init();
	locolor_act_init(TRUE);
	ctx = new_rgba_context();
	ctx->color_mode = color_mode_from_env();
	//ctx->debug = 1;
	//ctx->contrast_ratio = 3.0;

	/* The new context provides a palette that can contain named color
	 * definitions.  it is initially empty, but we can load new definitions
	 * into it like this: */
	printf("Loading user context palette '%s'...",userpalette);
	entries = load_colorchip_hash(userpalette,ctx);
	printf("found %d entries.\n",entries);

	printf("Loading user context styles '%s'...",userstyles);
	entries = load_style_hash(userstyles,ctx);
	printf("found %d entries.\n",entries);

	style_blendfn_t *fn[] = {
		style_blendfn_generic,
		style_blendfn_gamma,
		style_blendfn_hsva,
		style_blendfn_rainbow
	};

	char *blend[]={ 
		"UofM",
		"Purple",
		"Orange",
		"MSU"
	};

	for(int i=0;i<ARRAY_SIZE(blend);i++) {
		style = new_style();
		if(style_sscanf(blend[i],&style,ctx)) {
			sl = g_list_append(sl,style);
		} else {
			free_style(style);
		}
	}

	int steps = 25;
	for(int i=0;i<=steps;i++) {
		/* set s to the start of the buffer. */
		s = buf; 
		s += g_snprintf(s,eos-s,"Style Test %3.2f: ",(double)i/steps);
		/* fetch the style */	

		for(int f=0;f<ARRAY_SIZE(fn);f++) {
			st = style_blendfn_list((double)i/steps,sl,fn[f],ctx);
			s+=sgr_ansiStyle(s,eos-s,st,ctx);
			free_style(st);
			//s += g_snprintf(s,eos-s,"Lorem ipsum dolor sit amet, consectetur");
			s += g_snprintf(s,eos-s,"Lorem ipsum dolor ");
			/*put the output back to normal. */
			s+=sgr_ansiNormal(s,eos-s,ctx);
		}
		/* and print it all out. */
		puts(buf);
	}

	s = buf; 
	s += g_snprintf(s,eos-s,"Style Blend Text: ");
	s += style_blend_text(s,eos-s,"Taste the rainbow of blended text styles available!",sl,NULL,ctx);
	s+=sgr_ansiNormal(s,eos-s,ctx);
	puts(buf);

	/* now the easy way. */
	locolor_printf(ctx,"Act blend test: $blend('Lorem ipsum dolor sit amet, consectetur adipiscing elit','UofM') - that\'s a style.\n");
	locolor_printf(ctx,"Act blend test: $fg(Red)Red text, $blend('Lorem ipsum dolor sit amet, consectetur adipiscing elit','UofM') - that\'s a style.$reset\n");
	locolor_printf(ctx,"Act blend test: $blend('Lorem ipsum dolor sit amet, consectetur adipiscing elit','UofM',MSU)\n");
	locolor_printf(ctx,"Act blend test: $blend('Lorem ipsum dolor sit amet, consectetur adipiscing elit',Indigo,Orange)\n");

	g_list_free_full(sl,(GDestroyNotify)free_style);

	/* remember that you're responsible for throwing out anything you've
	 * allocated! */
	free_rgba_context(ctx);
	locolor_act_quit();
	locolor_quit();
}

int main(int argc, char** argv) {
	demo();
//	quickblends();
//	attrdemo();
//	styledemo();
	blendstyledemo();
	//gvp_test();
}

